import asyncio
from asyncio import AbstractEventLoop
from os import _exit
from typing import Callable, Dict

from pydantic import BaseModel

from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.utils.list_of_services import SERVICES


class AsyncioBroker(BrokerBase):
    __is_running: bool
    __is_active: bool

    __commands: Dict[str, Callable]

    # override
    def __init__(self, config: Config, logger=print):
        self.config = config
        self.__log = logger
        self.__is_running = False
        self.__is_active = False
        self.__brokers = {}

        self.__commands = {"start": self.start, "pause": self.pause, "stop": self.stop}

    # override
    def run(self):
        # self.__init_sockets(self.config)
        self.__is_running = True
        self.__is_active = True

    # override
    async def send(self, service: SERVICES, msg: str, *args):
        if self.__is_active:
            new_args = []
            # data = [service.value.encode('utf-8'), msg.encode('utf-8')]
            for arg in args:
                if isinstance(arg, BaseModel):
                    arg = arg.dict()
                    new_args.append(arg)
        #         arg = dumps(arg, default=str)
        #         data.append(arg.encode('utf-8'))
        #     self.__pub.send_multipart(data)

        service: AsyncioBroker = self.__find_service(service)
        # print('new args ', new_args)
        if len(new_args) > 0:
            await service.handle(msg, *new_args)
        else:
            await service.handle(msg, *args)
        await asyncio.sleep(0.00001)

    # @abstractmethod
    # def _handle_zmq_message(self, msg: str):
    #     pass

    # @abstractmethod
    # def _asyncio_loop(self, loop:asyncio.AbstractEventLoop):
    #     pass

    # def _loop(self):
    #     self._asyncio_loop(self.__loop)

    def __find_service(self, service: SERVICES):
        return self.__brokers[service.value]

    def register(self, command: str, func: Callable):
        self.__commands[command] = func

    def register_broker(self, service: SERVICES, service_object):
        self.__brokers[service.value] = service_object

    def create_listeners(self, loop: AbstractEventLoop):
        self.__log("Start main loop")
        # for sub in self.__subs:
        #     loop.create_task(self.__listen_zmq(sub))

    # async def __listen_zmq(self, sub):
    #     while self.__is_running:

    #         if self.__is_active:
    #             data = await sub.recv_multipart()
    #             if data:
    #                 await self.__handle(data)
    #             await asyncio.sleep(0.00001)
    #         else:
    #             await asyncio.sleep(0.01)
    #     self.__deinit()

    # def __init_sockets(self, config: Config):
    #     sub_context = zmq.asyncio.Context()
    #     pub_context = zmq.Context()
    #     for sub in config.sub:
    #         s = sub_socket(sub_context, 'tcp://' + config.ip + ':' + str(sub.port), sub.topic)
    #         self.__subs.append(s)
    #     self.__pub = pub_socket(pub_context, 'tcp://*:' + str(config.pub.port))

    async def handle(self, cmd, *args):
        func = self.__commands.get(cmd)
        if func:
            # self.__log(f'Receive "{cmd}" command')
            if args:
                await func(*args)
            else:
                await func()
        else:
            self.__log(f"Command '{cmd}' not registered")

    def stop(self):
        self.__log("Service stoped")
        self.__is_running = False
        _exit(1)

    def start(self):
        self.__log("Service started")
        self.__is_active = True

    def pause(self):
        self.__log("Service paused")
        self.__is_active = False


# def sub_socket(context: zmq.asyncio.Context(), url: str, topic: str):
#     sub = context.socket(zmq.SUB)
#     sub.connect(url)
#     sub.subscribe(topic)
#     return sub

# def pub_socket(context: zmq.Context(), url: str):
#     pub = context.socket(zmq.PUB)
#     ret = pub.bind(url)
#     return pub



from os import getenv
from typing import Union

import pandas as pd

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.data_sources.tradingview.downloader import (
    Interval as TRADINGVIEW_INTERVALS,
)
from backtesterrb30.libs.data_sources.tradingview.downloader import (
    TradingviewDownloader,
)
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.timestamps import datetime_to_timestamp


class TradingView(DataSource):
    INTERVALS = TRADINGVIEW_INTERVALS
    NAME = "tradingview"

    def __init__(self, logger=print):
        super().__init__(False, logger)
        trading_view_user = getenv("trading_view_user")
        trading_view_password = getenv("trading_view_password")
        if trading_view_user is None or trading_view_password is None:
            raise Exception(self.NAME + " No authorization heys provided")
        self.client = TradingviewDownloader(trading_view_user, trading_view_password)
        self.df_to_clip = None

    # override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        symbol, exchange, fut_contract = self.get_symbol_params(data.symbol)
        self.df_to_clip: pd.DataFrame = await self.client.get_hist(
            symbol,
            datetime_to_timestamp(data.backtest_date_start),
            exchange,
            data.interval,
            fut_contract,
        )
        return True

    # override
    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        self._log("Downloading tradingview data", instrument, interval)
        if time_stop is None:
            pass
            # TODO sprawdzic do czego byl ten kod
            # df_orig = df[df["timestamp"] >= time_start]
            # df = df_orig.iloc[:, [0, 1]]
            # milis = self._get_interval_miliseconds(interval)
            # if not isinstance(milis, int) or milis == 0:
            #     self._log("Warning, binance cannot get close price")
            # else:
            #     df_close = df_orig.iloc[-1:, [0, 4]]
            #     df_close.columns = df.columns
            # df = pd.concat([df, df_close])
        else:
            df = self.__clip_df(time_start, time_stop, self.df_to_clip)
        return df

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        interval_map = {
            "1": 60 * 1000,
            "3": 3 * 60 * 1000,
            "5": 5 * 60 * 1000,
            "15": 15 * 60 * 1000,
            "30": 30 * 60 * 1000,
            "45": 45 * 60 * 1000,
            "1H": 60 * 60 * 1000,
            "2H": 2 * 60 * 60 * 1000,
            "3H": 3 * 60 * 60 * 1000,
            "4H": 4 * 60 * 60 * 1000,
            "1D": 24 * 60 * 60 * 1000,
            "1W": 7 * 24 * 60 * 60 * 1000,
            "1M": 30 * 24 * 60 * 60 * 1000,  # Approximation
        }
        return interval_map.get(interval, None)

    def __clip_df(self, time_start: int, time_stop: int, df: pd.DataFrame):
        df = df[df["timestamp"] >= time_start]
        df = df[df["timestamp"] <= time_stop]
        df = df.iloc[:-1, [1, 2]]
        return df

    @staticmethod
    def create_symbol_name(symbol: str, exchange: str, fut_contract: str = ""):
        if not symbol or symbol == "":
            raise Exception("bad symbol name")
        if not exchange or exchange == "":
            raise Exception("bad exchange name")
        return (
            f"{symbol}_{exchange}_{fut_contract}"
            if fut_contract and fut_contract != ""
            else f"{symbol}_{exchange}"
        )

    @staticmethod
    def get_symbol_params(symbol: str):
        value = symbol.split("_")
        if len(value) < 2:
            raise Exception("Bad symbol")
        if len(value) == 2:
            value.append(None)
        return tuple(value)



from datetime import datetime, timezone


def datetime_to_timestamp(datetime: datetime):
    if datetime.tzinfo is None or datetime.tzinfo.utcoffset(datetime) is None:
        datetime = datetime.replace(tzinfo=timezone.utc)
    return int(datetime.timestamp() * 1000)


def timestamp_to_datetime(timestamp: int):
    return datetime.utcfromtimestamp(timestamp / 1000)



import asyncio
import time
from enum import Enum
from typing import Union

import pandas as pd
from pycoingecko import CoinGeckoAPI

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol

# from backtesterrb30.libs.utils.historical_sources import COINGECKO_INTERVALS
from backtesterrb30.libs.utils.timestamps import (
    datetime_to_timestamp,
    timestamp_to_datetime,
)


class COINGECKO_INTERVALS_2(Enum):
    day4 = "day4"


class CoingeckoDataSource(DataSource):
    INTERVALS = COINGECKO_INTERVALS_2
    NAME = "coingecko"

    def __init__(self, logger=print):
        super().__init__(False, logger)
        self.client = CoinGeckoAPI()
        self.__last_request_time = 0

    # override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        try:
            candles = self.client.get_coin_ohlc_by_id(
                data.symbol, vs_currency="usd", days="365"
            )
        except Exception as e:
            if e.args[0]["error"] == "Could not find coin with the given id":
                raise e
        df = pd.DataFrame(candles)
        if datetime_to_timestamp(data.backtest_date_start) < df.iloc[0, 0]:
            raise (
                Exception(
                    "Error. First avaliable date of ",
                    data.symbol,
                    "is",
                    timestamp_to_datetime(df.iloc[0, 0]),
                )
            )
        if datetime_to_timestamp(data.backtest_date_stop) > df.iloc[-1, 0]:
            raise Exception(
                "Error. Last avaliable date of ",
                data.symbol,
                "is",
                timestamp_to_datetime(df.iloc[0, 0]),
            )
        return True

    # override
    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        self._log("Downloading coingecko data", instrument)
        if interval == COINGECKO_INTERVALS_2.day4.value:
            actual_time = time.time()
            if actual_time - self.__last_request_time < 0.025:
                self._log("waitin")
                await asyncio.sleep(0.025 - (actual_time - self.__last_request_time))
            candles = self.client.get_coin_ohlc_by_id(
                instrument, vs_currency="usd", days="365"
            )
            df_orig = pd.DataFrame(candles)
            df = df_orig[df_orig[0] >= time_start]
            if time_stop is None:
                df = df_orig.iloc[:-1, [0, 1]]
            else:
                df = df[df[0] <= time_stop]
                df = df.iloc[:-1, [0, 1]]
        else:
            self._raise_error("No such interval")
        return df

    def get_current_price(self, data_symbol: DataSymbol):
        price = self.client.get_price(data_symbol.symbol, vs_currencies="USD")
        return price["polkadot"]["usd"]

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        if interval == COINGECKO_INTERVALS_2.day4.value:
            return 1000 * 60 * 60 * 24 * 4



import importlib.resources
import json
from datetime import datetime
from enum import Enum
from os import getcwd, remove, system, walk
from os.path import join
from shutil import rmtree
from typing import Union

import pandas as pd

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.timestamps import timestamp_to_datetime

# from backtesterrb30.historical_data_feeds.modules.utils import validate_dataframe_timestamps


class DUKASCOPY_INTERVALS_2(str, Enum):
    tick: str = "tick"
    minute: str = "minute"
    minute5: str = "minute5"
    minute15: str = "minute15"
    minute30: str = "minute30"
    hour: str = "hour"
    hour4: str = "hour4"
    day: str = "day"
    month: str = "month"


class DukascopyDataSource(DataSource):
    INTERVALS = DUKASCOPY_INTERVALS_2
    NAME = "dukascopy"

    def __init__(self, logger=print):
        super().__init__(False, logger)

    def __get_ducascopy_interval(self, interval: str) -> str:
        if interval == "tick":
            return "tick"
        if interval == "minute":
            return "m1"
        if interval == "minute5":
            return "m5"
        if interval == "minute15":
            return "m15"
        if interval == "minute30":
            return "m30"
        if interval == "hour":
            return "h1"
        if interval == "day":
            return "d1"
        if interval == "month":
            return "mn1"

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        if interval == "tick":
            return None
        if interval == "minute":
            return 60 * 1000
        if interval == "minute5":
            return 5 * 60 * 1000
        if interval == "minute15":
            return 15 * 60 * 1000
        if interval == "minute30":
            return 30 * 60 * 1000
        if interval == "hour":
            return 60 * 60 * 1000
        if interval == "day":
            return 24 * 60 * 60 * 1000
        if interval == "month":
            return None

    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        # https://raw.githubusercontent.com/Leo4815162342/dukascopy-node/master/src/utils/instrument-meta-data/generated/raw-meta-data-2022-04-23.json
        # response = requests.get("http://api.open-notify.org/astros.json")
        from_datetime_timestamp = int(
            round(datetime.timestamp(data.backtest_date_start) * 1000)
        )
        # f = open('temporary_ducascopy_list.json')
        # instrument_list = load(f)['instruments']
        with importlib.resources.open_text(
            "backtesterrb30", "temporary_ducascopy_list.json"
        ) as file:
            instrument_list = json.load(file)["instruments"]
        # validate if instrument exists:
        if data.symbol.upper() not in [
            v["historical_filename"] for k, v in instrument_list.items()
        ]:
            self._log(
                'Error. Instrument "' + data.symbol + '" does not exists on dukascopy.'
            )
            return False

        # validate it timestamps perios is right:
        for k, v in instrument_list.items():
            if v["historical_filename"] == data.symbol.upper():
                first_timestamp = int(v["history_start_day"])
                if first_timestamp > from_datetime_timestamp:
                    print(
                        "Error. First avaliable date of ",
                        data.symbol,
                        "is",
                        datetime.fromtimestamp(first_timestamp / 1000.0),
                    )
                    return False

        return True

    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        self._log("Downloading dukascopy data", instrument)
        """
        documentation: 
        https://github.com/Leo4815162342/dukascopy-node
        """

        duca_interval = self.__get_ducascopy_interval(interval)
        time_start_datetime = timestamp_to_datetime(time_start)
        if [
            time_start_datetime.hour,
            time_start_datetime.minute,
            time_start_datetime.second,
            time_start_datetime.microsecond,
        ] != [0, 0, 0, 0]:
            raise Exception(
                "Cannot dowload data with this time_start! Intervals must be in day graduality"
            )
        time_stop_datetime = timestamp_to_datetime(time_stop)
        if [
            time_stop_datetime.hour,
            time_stop_datetime.minute,
            time_stop_datetime.second,
            time_stop_datetime.microsecond,
        ] != [0, 0, 0, 0]:
            raise Exception(
                "Cannot dowload data with this time_stop! Intervals must be in day graduality"
            )
        from_param = datetime.fromtimestamp(time_start // 1000.0).strftime("%Y-%m-%d")
        to_param = datetime.fromtimestamp(time_stop // 1000.0).strftime("%Y-%m-%d")
        if from_param == to_param:
            raise Exception("The same start and stop date")
        here = getcwd()
        cache_path = join(here, "cache_dukascopy")
        try:
            rmtree(cache_path)
        except Exception:
            pass
        string_params = [
            " -i " + instrument,
            " -from " + from_param,
            " -to " + to_param,
            " -s",
            " -t " + duca_interval,
            " -fl",
            " -f csv",
            " -dir " + cache_path,
            " -p bid",
        ]
        command = "npx --yes dukascopy-node"
        for param in string_params:
            command += param
        result = system(command)
        if result != 0:
            self._log("Node is not installed?")
            raise Exception("Command not found or failed to execute: " + command)
        name_of_created_file = next(walk(cache_path), (None, None, []))[2][0]
        df = pd.read_csv(
            join(cache_path, name_of_created_file), index_col=None, header=None
        )
        if duca_interval == "tick":
            df = df.iloc[1:, [0, 2]]
        else:
            df = df.iloc[1:, [0, 1]]
        if name_of_created_file and name_of_created_file != "":
            remove(join(cache_path, name_of_created_file))
        # print('Dukascopy df shape after dl', df.shape)
        # print('head', str(df.head(1)))
        # print('tail', str(df.tail(1)))
        return df



import pandas as pd


def validate_dataframe_timestamps(
    df: pd.DataFrame, interval_step_miliseconds: int, time_start: int, time_stop: int
) -> pd.DataFrame:
    try:
        if interval_step_miliseconds > 1000 * 60 * 60 * 24:
            print("cannot validate df with this interval")
            return df
        timestamp_interval = interval_step_miliseconds
        list_timestamps = range(time_start, time_stop, timestamp_interval)
        for i, timestamp in enumerate(list_timestamps):
            if i == 0:
                continue
            count = 0
            while int(df.iloc[i][0]) != int(timestamp):
                count += 1
                df = pd.concat(
                    [df.iloc[:i], df.iloc[i - 1 : i], df.iloc[i:]], ignore_index=True
                )
                df.iloc[i, 0] = int(timestamp)
    except IndexError:
        pass
    return df



def singleton(class_):
    """
    Usage:

    @singleton
    class NameOfClass():
        pass
    """
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance



import asyncio
from typing import List

from backtesterrb30.historical_data_feeds.functions import (
    check_symbol_data_exists,
    create_downloading_clients,
)
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import (
    InstrumentFile,
)
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


class HistoricalDownloader:
    def __init__(
        self,
        data_schema: DataSchema,
        historical_sources_array,
        downloaded_data_path,
        logger=print,
    ):
        self._log = logger
        self.__data_schema: DataSchema = data_schema
        self.__historical_sources_array = historical_sources_array
        self.__data_sources = {}
        self.__data_to_download_2 = None
        self.__downloaded_data_path = downloaded_data_path

    def run(self, loop):
        self.__get_data_to_download(self.__data_schema)
        create_downloading_clients(
            self.__historical_sources_array,
            self.__data_schema,
            self.__data_sources,
            self._log,
        )
        self.__validate_data_schema_instruments(self.__data_schema.data, loop)
        self.__validate_symbols_to_download(self.__data_schema, loop)
        self.__download_prepared_data(self.__data_schema, loop)
        return self.__data_to_download_2

    def __get_data_to_download(self, data_schema: DataSchema):
        for data_symbol in data_schema.data:
            data_symbol.additional_properties["files_to_download"]: List[
                InstrumentFile
            ] = []
            data_symbol.additional_properties["files_to_download"] = (
                check_symbol_data_exists(data_symbol, self.__downloaded_data_path)
            )
            if self.__data_to_download_2 is None:
                self.__data_to_download_2 = []
            self.__data_to_download_2 = (
                self.__data_to_download_2
                + data_symbol.additional_properties["files_to_download"]
            )

    def __validate_symbols_to_download(
        self, data_schema: DataSchema, loop: asyncio.AbstractEventLoop
    ):
        self._log("validating symbols before download")
        for symbol in data_schema.data:
            if symbol.additional_properties["files_to_download"] != []:
                data_source_client: DataSource = self.__get_data_source_client(
                    symbol.historical_data_source
                )
                res = loop.run_until_complete(
                    data_source_client.validate_instrument(symbol)
                )
                if not res:
                    raise Exception("Error while validation symbol.")

    def __download_prepared_data(
        self, data_schema: DataSchema, loop: asyncio.AbstractEventLoop
    ):
        for symbol in data_schema.data:
            if symbol.additional_properties["files_to_download"] != []:
                loop.create_task(self.__download_symbol_data(symbol))

    def __validate_data_schema_instruments(
        self, data_symbol_array: List[DataSymbol], loop: asyncio.AbstractEventLoop
    ):
        self._log("Data_schema validation")
        # check for duplicates:
        seen = []
        for x in data_symbol_array:
            data = str(x.symbol) + str(x.historical_data_source)
            if data in seen:
                raise Exception("Duplicate data symbols")
            else:
                seen.append(data)
        # check other
        number_of_trigger_feeders = 0
        for data in data_symbol_array:
            if data.backtest_date_start is None:
                raise Exception(
                    'Error. You must provide "backtest_date_start" field in data_schema file while you are backtesting your strategy'
                )
            if data.backtest_date_start >= data.backtest_date_stop:
                raise Exception(
                    'Error. You have provided "backtest_date_start" is equal or bigger than "backtest_date_start" '
                )
            if [
                data.backtest_date_start.hour,
                data.backtest_date_start.minute,
                data.backtest_date_start.second,
                data.backtest_date_start.microsecond,
            ] != [0, 0, 0, 0]:
                raise Exception(
                    'Error. Provide your "backtest_date_start" and "backtest_date_stop" in a day accuracy like: "backtest_date_start": datetime(2020,6,1)'
                )
            if data.historical_data_source not in self.__historical_sources_array:
                raise Exception(
                    "Error. This historical_data_source not implemented yet"
                )
            if data.trigger_feed:
                number_of_trigger_feeders += 1
        if number_of_trigger_feeders < 1:
            raise Exception(
                'Error. Your "data_schema.py" must have at least one instrument that triggers feeds'
            )

    def __get_data_source_client(self, historical_source: str) -> DataSource:
        client = self.__data_sources[historical_source]
        if not client:
            raise Exception("Error, no registered source client")
        return client

    async def __download_symbol_data(self, symbol: DataSymbol):
        files_to_download: List[InstrumentFile] = symbol.additional_properties[
            "files_to_download"
        ]
        await self.__download_symbol_files(
            files_to_download, symbol.historical_data_source
        )

    async def __download_symbol_files(self, files: List[InstrumentFile], source: str):
        data_source_client: DataSource = self.__get_data_source_client(source)
        for file in files:
            await data_source_client.download_instrument(
                self.__downloaded_data_path, file
            )



from datetime import datetime
from os import listdir
from os.path import isfile, join
from typing import List

import pandas as pd

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import (
    InstrumentFile,
)
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


def synchronize_dataframes(list_of_dfs: List[dict], last_row: list) -> List[list]:
    rows = []
    c = 0
    while True:
        c += 1
        row = []
        feeding_next_timestamps = [
            df["actual_raw"][0]
            for df in list_of_dfs
            if df["trigger_feed"] and not df["consumed"]
        ]
        if len(feeding_next_timestamps) == 0:
            break
        min_timestamp = min(feeding_next_timestamps)
        row.append(int(min_timestamp))
        for df_obj in list_of_dfs:
            while True:
                if df_obj["actual_raw"][0] > min_timestamp:
                    row.append(df_obj["prev_raw"][1])
                    break
                elif df_obj["actual_raw"][0] == min_timestamp:
                    row.append(df_obj["actual_raw"][1])
                    try:
                        df_obj["prev_raw"] = df_obj["actual_raw"]
                        i, v = next(df_obj["rows_iterator"])
                        df_obj["actual_raw"] = list(v)
                    except StopIteration:
                        df_obj["consumed"] = True
                    break
                else:
                    try:
                        df_obj["prev_raw"] = df_obj["actual_raw"]
                        i, v = next(df_obj["rows_iterator"])
                        df_obj["actual_raw"] = list(v)
                    except StopIteration:
                        row.append(df_obj["actual_raw"][1])
                        df_obj["consumed"] = True
                        break
        if len(last_row) == 0 or row[0] != last_row[0]:
            rows.append(row)
            # print(row)
    return rows


def get_instrument_files(symbol: DataSymbol) -> List[str]:
    instrument_files: List[InstrumentFile] = []
    params_touple = (symbol.historical_data_source, symbol.symbol, symbol.interval)

    if symbol.backtest_date_start.year < symbol.backtest_date_stop.year:
        instrument_files.append(
            InstrumentFile.from_params(
                *params_touple,
                symbol.backtest_date_start,
                datetime(symbol.backtest_date_start.year + 1, 1, 1),
            )
        )

        for i in range(
            symbol.backtest_date_stop.year - symbol.backtest_date_start.year - 1
        ):
            instrument_files.append(
                InstrumentFile.from_params(
                    *params_touple,
                    datetime(symbol.backtest_date_start.year + i + 1, 1, 1),
                    datetime(symbol.backtest_date_start.year + i + 2, 1, 1),
                )
            )

        instrument_files.append(
            InstrumentFile.from_params(
                *params_touple,
                datetime(symbol.backtest_date_stop.year, 1, 1),
                symbol.backtest_date_stop,
            )
        )
    else:
        instrument_files.append(
            InstrumentFile.from_params(
                *params_touple, symbol.backtest_date_start, symbol.backtest_date_stop
            )
        )
    return instrument_files


def create_downloading_clients(
    historical_sources_array: list,
    data_schema: DataSchema,
    data_sources: dict,
    log=print,
):
    for source in historical_sources_array:
        if source in [
            data.historical_data_source
            for data in data_schema.data
            if data.additional_properties["files_to_download"] != []
        ]:
            data_sources[source]: DataSource = getattr(HISTORICAL_SOURCES, source)(log)


def check_symbol_data_exists(
    data_symbol: DataSymbol, downloaded_data_path: str
) -> List[InstrumentFile]:
    """
    data scheme
    <symbol>__<source>__<interval>__<date-from>__<date-to>
    all instruments are downloaded in year files.
    """
    files: List[InstrumentFile] = get_instrument_files(data_symbol)
    # self._log('files to download "'+str(data_symbol.symbol)+'" :', files)
    files_in_directory = [
        f
        for f in listdir(downloaded_data_path)
        if isfile(join(downloaded_data_path, f))
    ]
    files_to_download = list(
        set([f.to_filename() for f in files]) - set(files_in_directory)
    )
    files_to_download = [
        InstrumentFile.from_filename(file) for file in files_to_download
    ]
    return files_to_download


def map_raw_to_instruments(raw: list, instruments: list):
    last_raw_obj = {}
    if len(raw) != len(instruments):
        print("raw", raw, "iinstruments", instruments)
        raise Exception(
            "Error in map_raw_to_instruments. Lengths of raw and list of instruments are not equal"
        )
    for value, instrument in zip(raw, instruments):
        last_raw_obj[instrument] = value
    return last_raw_obj


def load_files_array(
    downloaded_data_path, files_array: List[InstrumentFile]
) -> List[pd.DataFrame]:
    loaded_files = []
    for file in files_array:
        file_name = file.to_filename()
        columns = ["timestamp", file.identifier]
        df = pd.read_csv(
            join(downloaded_data_path, file_name),
            index_col=None,
            header=None,
            names=columns,
        )
        loaded_files.append(df)
    return loaded_files


def prepare_dataframes_to_synchronize_2(
    data_schema: DataSchema,
    all_columns: list,
    last_row: list,
    files_array_loaded: List[pd.DataFrame],
) -> List[dict]:
    list_of_dfs = []
    for data_element in data_schema.data:
        cols = ["timestamp", data_element.identifier]
        actual_raw = [0, 0]
        prev_raw = [0, 0]
        df = pd.DataFrame([], columns=cols)
        for element in files_array_loaded:
            element: pd.DataFrame = element
            if len(element.columns) != 2:
                raise Exception("Bad dataframe. Loaded dataframe should have 2 columns")
            if data_element.identifier == element.columns[1]:
                df = element
        if not df.empty and last_row != []:
            last_raw_mapped = map_raw_to_instruments(last_row, all_columns)
            prev_raw[0] = last_raw_mapped["timestamp"]
            prev_raw[1] = last_raw_mapped[data_element.symbol]
        obj = {
            "trigger_feed": data_element.trigger_feed,
            "rows_iterator": df.iterrows(),
            "actual_raw": actual_raw,
            "prev_raw": prev_raw,
            "consumed": False,
        }
        # prepare to load:
        if obj["actual_raw"][0] == 0:
            try:
                i, v = next(obj["rows_iterator"])
                obj["actual_raw"] = list(v)
            except StopIteration:
                obj["consumed"] = True
        list_of_dfs.append(obj)
    return list_of_dfs


def load_data_frame_ticks_2(
    data_schema: DataSchema,
    columns: list,
    downloaded_data_path: str,
    last_row: list,
    files_array: List[InstrumentFile],
) -> List[list]:
    """
    Function is geting files array from one period that are going to be loaded.
    Function returns synchronized data in list of lists which are ready to send to engine.
    """
    files_loaded = load_files_array(downloaded_data_path, files_array)
    list_of_dfs = prepare_dataframes_to_synchronize_2(
        data_schema, columns, last_row, files_loaded
    )
    rows = synchronize_dataframes(list_of_dfs, last_row)
    return rows


def load_data_frame_with_dfs(
    data_schema: DataSchema,
    columns: list,
    last_row: list,
    dataframes_array: List[pd.DataFrame],
) -> List[list]:
    list_of_dfs = prepare_dataframes_to_synchronize_2(
        data_schema, columns, last_row, dataframes_array
    )
    rows = synchronize_dataframes(list_of_dfs, last_row)
    return rows



from typing import List, Optional

from pydantic import BaseModel

from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import (
    CustomChartElement,
)


class CustomChart(BaseModel):
    name: str
    display_on_price_chart: Optional[bool] = False
    log_scale: Optional[bool] = False
    color: Optional[str] = "orange"
    chart: List[CustomChartElement]



from typing import List

from pydantic import BaseModel


class Position(BaseModel):
    number_of_actions: float = 0
    position_outcome: float = 0
    buy_summary_cost: float = 0
    sell_summary_cost: float = 0
    last_instrument_price: float = 0
    trades: List = []



from datetime import datetime
from enum import Enum
from os import getenv
from typing import Union

import pandas as pd
from binance import Client
from binance.helpers import interval_to_milliseconds

from backtesterrb30.libs.data_sources.data_source_base import DataSource

# from backtesterrb30.libs.utils.historical_sources import BINANCE_INTERVALS
# from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


class BINANCE_INTERVALS_2(str, Enum):
    tick: str = "tick"
    minute: str = "minute"
    minute3: str = "minute3"
    minute5: str = "minute5"
    minute15: str = "minute15"
    minute30: str = "minute30"
    hour: str = "hour"
    hour2: str = "hour2"
    hour4: str = "hour4"
    hour6: str = "hour6"
    hour8: str = "hour8"
    hour12: str = "hour12"
    day: str = "day"
    day3: str = "day3"
    week: str = "week"
    month: str = "month"


class BinanceDataSource(DataSource):
    INTERVALS = BINANCE_INTERVALS_2
    NAME = "binance"

    def __init__(self, logger=print):
        super().__init__(False, logger)
        binance_api_secret = getenv("binance_api_secret")
        binance_api_key = getenv("binance_api_key")
        if binance_api_secret is None or binance_api_key is None:
            raise Exception(self.NAME + " No authorization heys provided")
        self.client = Client(binance_api_key, binance_api_secret)
        self._test_counter = 0

    # override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        # validate if instrument exists:
        exchange_info = self.client.get_exchange_info()
        if data.symbol not in [s["symbol"] for s in exchange_info["symbols"]]:
            self._log(
                'Error. Instrument "' + data.symbol + '" does not exists on binance.'
            )
            return False
        # validate it timestamps perios is right:
        from_datetime_timestamp = int(
            round(datetime.timestamp(data.backtest_date_start) * 1000)
        )
        if data.interval == BINANCE_INTERVALS_2.tick:
            binance_interval = Client.KLINE_INTERVAL_1MINUTE
        else:
            binance_interval = self.__get_binance_interval(data.interval.value)
        first_timestamp = self.client._get_earliest_valid_timestamp(
            data.symbol, binance_interval
        )
        if first_timestamp > from_datetime_timestamp:
            self._log(
                "Error. First avaliable date of ",
                data.symbol,
                "is",
                datetime.utcfromtimestamp(first_timestamp / 1000.0),
            )
            return False
        return True

    # override
    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        self._log("Downloading binance data", instrument)
        if interval == "tick":
            interval_timestamp = 1000 * 60 * 60
            trades_arr = []
            start_hour = time_start
            stop_hour = start_hour + interval_timestamp
            while stop_hour < time_stop:
                trades = self.client.get_aggregate_trades(
                    symbol=instrument, startTime=start_hour, endTime=stop_hour
                )
                # print('trades', trades)
                trades_arr = trades_arr + trades
                start_hour += interval_timestamp
                stop_hour += interval_timestamp
            df = pd.DataFrame(trades_arr)
            df = df[["T", "p"]]
        else:
            binance_interval = self.__get_binance_interval(interval)
            klines = self.client.get_historical_klines(
                instrument, binance_interval, time_start, time_stop
            )
            df_orig = pd.DataFrame(klines)
            df = df_orig.iloc[:-1, [0, 1]]
            if time_stop is None:
                df = df_orig.iloc[:, [0, 1]]
                milis = self._get_interval_miliseconds(interval)
                if not isinstance(milis, int) or milis == 0:
                    self._log("Warning, binance cannot get close price")
                else:
                    df_close = df_orig.iloc[-1:, [0, 4]]
                    df_close.columns = df.columns
                df = pd.concat([df, df_close])

            # if interval != STRATEGY_INTERVALS.tick.value:
            #     df = validate_dataframe_timestamps(df, interval, time_start, time_stop)

            # print('binance shape after dl: ', df.shape)
            # print('head', str(df.head(1)))
            # print('tail', str(df.tail(1)))

        return df

    def get_current_price(self, symbol: DataSymbol):
        # self._test_counter += 1
        # print('binance get curren price', self._test_counter)
        price = self.client.get_symbol_ticker(symbol=symbol.symbol)
        return price["price"]

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        return interval_to_milliseconds(self.__get_binance_interval(interval))

    def __get_binance_interval(self, interval: str) -> str:
        if interval == "minute":
            return Client.KLINE_INTERVAL_1MINUTE
        if interval == "minute3":
            return Client.KLINE_INTERVAL_3MINUTE
        if interval == "minute5":
            return Client.KLINE_INTERVAL_5MINUTE
        if interval == "minute15":
            return Client.KLINE_INTERVAL_15MINUTE
        if interval == "minute30":
            return Client.KLINE_INTERVAL_30MINUTE
        if interval == "hour":
            return Client.KLINE_INTERVAL_1HOUR
        if interval == "hour2":
            return Client.KLINE_INTERVAL_2HOUR
        if interval == "hour4":
            return Client.KLINE_INTERVAL_4HOUR
        if interval == "hour6":
            return Client.KLINE_INTERVAL_6HOUR
        if interval == "hour8":
            return Client.KLINE_INTERVAL_8HOUR
        if interval == "hour12":
            return Client.KLINE_INTERVAL_12HOUR
        if interval == "day":
            return Client.KLINE_INTERVAL_1DAY
        if interval == "day3":
            return Client.KLINE_INTERVAL_3DAY
        if interval == "week":
            return Client.KLINE_INTERVAL_1WEEK
        if interval == "month":
            return Client.KLINE_INTERVAL_1MONTH



import os

from backtesterrb30.libs.utils.list_of_services import SERVICES_ARRAY

serve_file_commands_array = [
    "from sys import argv\n",
    "from dotenv import load_dotenv\n",
    'load_dotenv(".env")\n',
    'load_dotenv(".additional_configs")\n',
    "microservice_name = argv[1]\n",
    'if microservice_name == "python_backtester":\n',
    "    from backtesterrb30.historical_data_feeds import run\n",
    'elif microservice_name == "python_engine":\n',
    "    from backtesterrb30.python_backtester import run\n",
    'elif microservice_name == "python_executor":\n',
    "    from backtesterrb30.python_engine import run\n",
    'elif microservice_name == "historical_data_feeds":\n',
    "    from backtesterrb30.python_executor import run\n"
    'elif microservice_name == "live_data_feeds":\n',
    "    from backtesterrb30.live_data_feeds import run\n",
]

run_file_commands_array = [
    "sudo -v\n"
    "trap ctrl_c INT\n"
    "function ctrl_c() {\n"
    '        echo "** Trapped CTRL-C"\n'
    "        sudo killall python3 \n"
    "        exit 2\n"
    "}\n"
    "sudo python3 serve.py python_backtester &\n"
    "sudo python3 serve.py python_engine &\n"
    "sudo python3 serve.py python_executor &\n"
    "sudo python3 serve.py historical_data_feeds &\n"
    "sudo python3 serve.py live_data_feeds &\n"
    "while true\n"
    "do\n"
    "    sleep 10\n"
    "done\n"
]


def run_all_microservices(
    strategy_path,
    backtest_state,
    strategy_file,
    model_class_name,
    executor_class_name,
    data_class_name,
):
    serve_prepared = False
    run_prepared = False
    for fname in os.listdir("."):
        if fname == "serve.py":
            serve_prepared = True
        if fname == "run.sh":
            run_prepared = True
    if not serve_prepared:
        with open("serve.py", "w") as f:
            for line in serve_file_commands_array:
                f.write(line)
    if not run_prepared:
        with open("run.sh", "w") as f:
            for line in run_file_commands_array:
                f.write(line)

    print("strategy_file", strategy_file)
    services_array = SERVICES_ARRAY

    def is_port_in_use(port: int) -> bool:
        import socket

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(("localhost", port)) == 0

    def create_port_configurations():
        start_port = 1000
        assigned = False
        while not assigned:
            port_in_use = False
            for port in range(start_port, start_port + 50):
                if is_port_in_use(port):
                    print(
                        "Port " + str(port) + " is already in use, changing port range."
                    )
                    start_port += 50
                    port_in_use = True
                    break
            if not port_in_use:
                assigned = True

        with open(".additional_configs", "w") as f:
            for i, service in enumerate(services_array):
                f.write(service + "_pubs=" + str(start_port + i) + "\n")
                subs_str = service + "_subs="
                for j in range(len(services_array)):
                    if j != i:
                        subs_str += str(start_port + j) + ","
                subs_str = subs_str[:-1]
                subs_str += "\n"
                f.write(subs_str)

    print("preparing microservice ports configuration")
    create_port_configurations()
    bck_state = "true" if backtest_state else "false"
    with open(".additional_configs", "a") as f:
        f.write("backtest_state=" + bck_state + "\n")
        f.write("STRATEGY_FILE=" + strategy_file + "\n")
        f.write("STRATEGY_PATH=" + strategy_path + "\n")
        f.write("MODEL_CLASS_NAME=" + model_class_name + "\n")
        f.write("EXECUTOR_CLASS_NAME=" + executor_class_name + "\n")
        f.write("DATA_CLASS_NAME=" + data_class_name + "\n")

    if backtest_state:
        os.system("bash run.sh")
    else:
        print("live strategies not implemented")

    # print('starting all containers')
    # system('docker-compose up --build --remove-orphans')



from typing import Union

# from backtesterrb30.historical_data_feeds.modules.utils import validate_dataframe_timestamps
import pandas as pd

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


class RB30DataSource(DataSource):
    INTERVALS = {}
    NAME = "rb30disk"

    def __init__(self, logger=print):
        super().__init__(False, logger)

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        pass

    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        return True

    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        return True



from typing import List, Optional

from pydantic import BaseModel

from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


class DataSchema(BaseModel):
    log_scale_valuation_chart: Optional[bool] = True
    data: List[DataSymbol]



# from enum import Enum
# from typing import Union

# class DUKASCOPY_INTERVALS(str, Enum):
#     tick: str='tick'
#     minute: str='minute'
#     minute5: str='minute5'
#     minute15: str='minute15'
#     minute30: str='minute30'
#     hour: str='hour'
#     hour4: str='hour4'
#     day: str='day'
#     month: str='month'

# class BINANCE_INTERVALS(str, Enum):
#     tick: str='tick'
#     minute: str='minute'
#     minute3: str='minute3'
#     minute5: str='minute5'
#     minute15: str='minute15'
#     minute30: str='minute30'
#     hour: str='hour'
#     hour2: str='hour2'
#     hour4: str='hour4'
#     hour6: str='hour6'
#     hour8: str='hour8'
#     hour12: str='hour12'
#     day: str='day'
#     day3: str='day3'
#     week: str='week'
#     month: str='month'


# class EXANTE_INTERVALS(str, Enum):
#     tick: str='tick'
#     minute: str='minute'
#     minute5: str='minute5'
#     minute30: str='minute30'
#     hour: str='hour'
#     day: str='day'


# class COINGECKO_INTERVALS(Enum):
#     day4='day4'

# @classmethod
# def get_small_intervals(cls):
#     return []


# class HISTORICAL_SOURCES(str, Enum):
#     binance: str='binance'
#     ducascopy: str='ducascopy'
#     rb30disk: str='rb30disk'
#     exante: str='exante'
#     coingecko: str='coingecko'

# HISTORICAL_INTERVALS_UNION = Union[BINANCE_INTERVALS, DUKASCOPY_INTERVALS, EXANTE_INTERVALS, COINGECKO_INTERVALS]



import asyncio
from abc import abstractmethod
from os.path import join
from typing import Union

import pandas as pd

from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import (
    InstrumentFile,
)
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol


class DataSource:
    def __init__(self, allow_synchronous_download: bool, logger=print):
        self._allow_synchronous_download = allow_synchronous_download
        self._log = logger

    def _raise_error(self, error_message):
        raise Exception(error_message)

    def __validate_dataframe(self, df: pd.DataFrame) -> bool:
        # if int(df.head(1)[0]) != expected_first_data_timestamp: return False
        # if int(df.tail(1)[0]) != expected_last_data_timestamp: return False
        # if len(df) != expected_data_length: return False
        if df.shape[1] != 2:
            raise Exception(
                "Bad shape of returned dataframe in data source " + self.NAME
            )
        return True

    async def download_instrument(
        self, downloaded_data_path: str, instrument_file: InstrumentFile
    ):
        await asyncio.sleep(0.1)
        df: pd.DataFrame = await self._download_instrument_data(
            instrument_file.instrument,
            instrument_file.interval,
            instrument_file.time_start,
            instrument_file.time_stop,
        )
        self.__validate_dataframe(df)
        df.to_csv(
            join(downloaded_data_path, instrument_file.to_filename()),
            index=False,
            header=False,
        )

    async def download_dataframe(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        await asyncio.sleep(0.1)
        df: pd.DataFrame = await self._download_instrument_data(
            instrument, interval, time_start, time_stop
        )
        self.__validate_dataframe(df)
        return df

    async def validate_instrument(self, data: DataSymbol) -> bool:
        await asyncio.sleep(0.1)
        if not isinstance(data.interval, self.INTERVALS):
            raise Exception(
                "Bad interval type in data symbol",
                data.symbol,
                data.historical_data_source,
                "check data_schema.py",
            )
        return await self._validate_instrument_data(data)

    @abstractmethod
    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        """
        This function should return number of miliseconds delay for provided interval
        """

    @abstractmethod
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        """
        Function should validate if data for shis DataSymbol is possible to download.
        It should check if
        - data.symbol exists
        - data.interval fot this data.symbol exists
        - timestamps range from 'data.backtest_date_start' to 'data.backtest_date_stop'
            for this data.symbol and data.interval exists
        """

    @abstractmethod
    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        """
        First downloaded timestamp should exactly exuals instrument_file.time_start (can be bigger if u are downloading ticks)
        Last downloaded timestamp should exactly exuals one interval before instrument_file.time_stop ()
        """

    def get_current_price(self, data: DataSymbol):
        """
        Function should return current price of provided data symbol
        """



from typing import List, Optional

from pydantic import BaseModel

from backtesterrb30.libs.interfaces.python_backtester.custom_chart import CustomChart


class DataFinish(BaseModel):
    # main_instrument_price: float
    custom_charts: Optional[List[CustomChart]] = None



from datetime import datetime
from enum import Enum
from typing import Any, Optional

from backtesterrb30.libs.interfaces.utils.custom_base_model import CustomBaseModel


class DataSymbol(CustomBaseModel):
    symbol: str
    # historical_data_source: HISTORICAL_SOURCES
    historical_data_source: str
    # interval: HISTORICAL_INTERVALS_UNION
    interval: Enum
    backtest_date_start: datetime
    backtest_date_stop: datetime
    trigger_feed: bool = True
    with_volume: bool = False
    display_chart_in_summary: bool = True
    custom_name: Optional[str]
    custom_data: Optional[Any]

    @property
    def identifier(self):
        return self.historical_data_source + "_" + self.symbol

    def get_buffer(self) -> list:
        if "buffer" in self.additional_properties:
            return self.additional_properties["buffer"]
        else:
            raise Exception(
                "No buffer registered, buffer avaliable only in Model class"
            )

    class Config:
        arbitrary_types_allowed = True



import asyncio
import time as tm
from datetime import datetime
from os.path import join
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from appdirs import user_cache_dir
from matplotlib.ticker import ScalarFormatter

from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.data_finish import (
    CustomChart,
    DataFinish,
)
from backtesterrb30.libs.interfaces.python_backtester.data_start import DataStart
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import (
    DebugBreakpoint,
)
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_backtester.positions import Position
from backtesterrb30.libs.interfaces.python_backtester.trade import Trade
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.list_of_services import SERVICES, SERVICES_ARRAY
from backtesterrb30.libs.utils.service import Service


class Backtester(Service):
    downloaded_data_path = user_cache_dir("rb30_cache")
    _broker: BrokerBase

    def __init__(
        self, config: Config, data_schema: DataSchema, loop=None, logger=print
    ):
        super().__init__(config, logger)
        self.config: Config = config
        self.__loop = loop
        self.__custom_event_loop = False
        if self.__loop is None:
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop = True
        self.data_schema: DataSchema = data_schema
        for dat in self.data_schema.data:
            dat.additional_properties["position"] = None
            dat.additional_properties["chart_data_frame"] = None

        # self.trading_instruments_charts = []
        self.cumulated_money_chart = []
        # self.positions: List[Position] = []
        self.__last_feed: LastFeed = {}
        self.__last_timestamp: int = 0

        self.__backtest_start_time = 0
        self.__fig = None
        self.__chart_displayed = False

    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._create_listeners(loop)
    #     loop.create_task(self.__update_chart())

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        self.__loop.create_task(self.__update_chart())
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("trade", self.__trade_event)
        self._broker.register("data_finish", self.__data_finish_event)
        self._broker.register("close_all_trades", self.__close_all_trades_event)
        self._broker.register("data_start", self.__data_start_event)
        self._broker.register("debug_breakpoint", self.__debug_breakpoint_event)
        self._broker.register("last_feed", self.__last_feed_event)

    # def _trigger_event(self, event):
    #     pass

    async def __stop_all_services(self):
        for service in SERVICES_ARRAY:
            if service != self.name:
                await self._broker.send(getattr(SERVICES, service), "stop")
        self._broker.stop()

    def __axis_format(self, ax, title):
        ax.set_title(title)
        ax.set_axisbelow(True)
        ax.yaxis.grid(color="gray", linestyle="dashed")

    async def __print_charts(self, custom_charts: List[CustomChart]):
        plt.close()

        # plot trades chart
        # position: Position = self.positions[0]
        number_of_custom_charts = 0
        if custom_charts is not None:
            number_of_custom_charts = len(
                [ch for ch in custom_charts if not ch.display_on_price_chart]
            )

        chartable_symbols = [
            sym for sym in self.data_schema.data if sym.display_chart_in_summary
        ]
        axs_number_used = 0

        # prepare axes
        self.__fig, axs = plt.subplots(
            nrows=1 + len(chartable_symbols) + number_of_custom_charts,
            ncols=1,
            sharex=True,
            figsize=(13, 13),
        )
        if not isinstance(axs, np.ndarray):
            # if only 1 axis its returned not in array
            axs = [axs]
        ax = None
        # plot instrment charts
        for sym in chartable_symbols:
            position: Position = sym.additional_properties["position"]
            # if position:
            if self.__last_timestamp is not None:
                main_chart = sym.additional_properties["chart_data_frame"].loc[
                    sym.additional_properties["chart_data_frame"]["timestamp"]
                    <= self.__last_timestamp
                ]
            ax = main_chart.plot(
                x="timestamp", y="price", kind="line", ax=axs[axs_number_used]
            )
            axs_number_used += 1
            self.__axis_format(ax, sym.symbol + " - " + sym.historical_data_source)
            if self.data_schema.log_scale_valuation_chart:
                ax.yaxis.set_major_formatter(ScalarFormatter())
            if position:
                normalized_quants = self.__normalize(
                    [abs(trade[2]) for trade in position.trades], (5, 15)
                )
                for trade, quant in zip(position.trades, normalized_quants):
                    ax.plot(
                        trade[0], trade[1], ".g" if trade[2] > 0 else ".r", ms=quant
                    )

        if len(self.cumulated_money_chart) > 0:
            # plot money chart
            money_df = pd.DataFrame(
                self.cumulated_money_chart, columns=["timestamp", "income"]
            )
            ax = money_df.plot(
                x="timestamp",
                y="income",
                kind="line",
                ax=axs[axs_number_used],
                sharex=ax,
            )
            axs_number_used += 1

        # plot custom charts
        if custom_charts is not None:
            for i, ch in enumerate(custom_charts):
                chart = [[c.timestamp, c.value] for c in ch.chart]
                custom_df = pd.DataFrame(chart, columns=["timestamp", ch.name])
                if ch.display_on_price_chart:
                    custom_df.plot(
                        x="timestamp",
                        y=ch.name,
                        kind="line",
                        ax=axs[0],
                        sharex=ax,
                        color=ch.color,
                    )
                else:
                    ax = custom_df.plot(
                        x="timestamp",
                        y=ch.name,
                        kind="line",
                        ax=axs[axs_number_used],
                        sharex=ax,
                        color=ch.color,
                    )
                    axs_number_used += 1
                    self.__axis_format(ax, ch.name)
                    if ch.log_scale:
                        ax.yaxis.set_major_formatter(ScalarFormatter())
        if not self.__chart_displayed:
            plt.ion()
            plt.show(block=False)
            self.__chart_displayed = True

    def __add_position(self, data_symbol: DataSymbol) -> Position:
        pos = Position()
        data_symbol.additional_properties["position"] = pos
        self.__update_last_instrument_prices()
        return pos

    async def __update_chart(self):
        while True:
            if self.__fig is not None:
                try:
                    self.__fig.canvas.draw()
                    self.__fig.canvas.flush_events()
                except Exception:
                    pass
            await asyncio.sleep(0.1)

    def __normalize(
        self, x, newRange=(0, 1)
    ):  # x is an array. Default range is between zero and one
        if len(x) == 0:
            return []
        xmin, xmax = np.min(x), np.max(x)  # get max and min from input array
        if xmin != xmax:
            norm = (x - xmin) / (xmax - xmin)  # scale between zero and one

            if newRange == (0, 1):
                return norm  # wanted range is the same as norm
            elif newRange != (0, 1):
                return (
                    norm * (newRange[1] - newRange[0]) + newRange[0]
                )  # scale to a different range.
            # add other conditions here. For example, an error messag
        return np.ones_like(x) * 15

    async def __recalculate_positions(self):
        positions: List[Position] = [
            elem.additional_properties["position"] for elem in self.data_schema.data
        ]
        positions = [pos for pos in positions if pos is not None]
        for pos in positions:
            pos.position_outcome = (
                -pos.buy_summary_cost
                - pos.sell_summary_cost
                + pos.number_of_actions * pos.last_instrument_price
            )
        current_capital = sum([pos.position_outcome for pos in positions])
        current_invested = sum(
            [
                abs(pos.number_of_actions) * pos.last_instrument_price
                for pos in positions
            ]
        )
        self.cumulated_money_chart.append([self.__last_timestamp, current_capital])
        await self._broker.send(
            SERVICES.python_executor, "set_current_capital_event", current_capital
        )
        await self._broker.send(
            SERVICES.python_executor, "set_current_invested_event", current_invested
        )

    def __get_data_symbol(self, symbol: str, source: str):
        arr = [
            sym
            for sym in self.data_schema.data
            if sym.symbol == symbol and sym.historical_data_source == source
        ]
        if len(arr) == 0:
            raise Exception("No data symbol found")
        if len(arr) > 1:
            raise Exception("More than one symbol found")
        return arr[0]

    async def __trade(self, trade: Trade):
        symbol: DataSymbol = self.__get_data_symbol(trade.symbol, trade.source)
        position = symbol.additional_properties["position"]
        if not position:
            position = self.__add_position(symbol)
        if trade.price == 0 and trade.timestamp == 0:
            trade.price = position.last_instrument_price
            trade.timestamp = self.__last_timestamp
        if trade.price <= 0 or trade.timestamp <= 0:
            self._log(
                "Cannot make trade with this price and timestamp:",
                trade.price,
                ";",
                trade.timestamp,
            )
            return
        self._log(
            f"Trade for [{symbol.symbol}]: time={datetime.utcfromtimestamp(trade.timestamp/1000)}, | value={trade.value}, price={trade.price}"
        )
        position.trades.append([trade.timestamp, trade.price, trade.value])
        # print('trade price', trade.price)
        position.number_of_actions += trade.value / trade.price
        if trade.value > 0:
            position.buy_summary_cost += trade.value
        else:
            position.sell_summary_cost += trade.value
        await self.__recalculate_positions()

    async def __print_summary(
        self,
        custom_charts: List[CustomChart],
        display_charts=True,
        # file_names: List[str],
        # main_instrument_price: float,
        # last_timestamp: Union[int, None] = None
    ):
        self._log("==========================")
        self._log(
            ("BREAKPOINT " if self.__last_timestamp is not None else "") + "SUMMARY"
        )
        finish_time = tm.time()
        time_of_backtest = finish_time - self.__backtest_start_time
        self._log("time of backtest:", round(time_of_backtest, 2), "[s]")
        # self._log('number of trades:', len(self.trades))
        # self._log('buy_summary_cost:', self.buy_summary_cost)
        # self._log('sell_summary_cost:', self.sell_summary_cost)
        # self._log('number of unrealized actions:', self.number_of_actions)
        # self._log('biggest investment: ', self.biggest_investment)
        # self._log('actual price:', main_instrument_price)
        # income = - self.buy_summary_cost - self.sell_summary_cost + self.number_of_actions * main_instrument_price
        if len(self.cumulated_money_chart) > 0:
            self._log("income:", self.cumulated_money_chart[-1][1])
        self._log("==========================")
        self._log("")
        if display_charts:
            await self.__print_charts(custom_charts)

    def __load_instrument_chart(instrument):
        loaded_df = None

        return loaded_df

    # COMMANDS

    async def __data_finish_event(self, finish_params):
        finish_params = DataFinish(**finish_params)
        self._log("")
        self._log("====================================================")
        self._log("BACKTEST FINISHED")
        await self.__print_summary(finish_params.custom_charts)
        self._log("====================================================")
        plt.show(block=True)
        await self.__stop_all_services()

    async def __trade_event(self, msg):
        trade: Trade = Trade(**msg)
        await self.__trade(trade)

    async def __last_feed_event(self, msg):
        # print('last feed event')
        self.__last_feed = LastFeed(**msg)
        self.__last_timestamp = self.__last_feed.last_feed[0]
        self.__update_last_instrument_prices()

    def __update_last_instrument_prices(self):
        for data_symbol, last_feed in zip(
            self.data_schema.data, self.__last_feed.last_feed[1:]
        ):
            position: Position = data_symbol.additional_properties["position"]
            if position:
                if isinstance(last_feed, (float, int)):
                    position.last_instrument_price = last_feed
                    continue
                if isinstance(last_feed, list):
                    print("feed is 3d")
                    if isinstance(last_feed[0], (float, int)):
                        position.last_instrument_price = last_feed[0]
                    continue

                raise Exception("Error during setting last instrument price")

    async def __close_all_trades_event(self):
        self._log("Received close all trades command NOT IMPLEMENTED")

        # TODO

    async def __data_start_event(self, start_params):
        start_params = DataStart(**start_params)
        self.__backtest_start_time = start_params.start_time

        for element in start_params.file_names:
            dfs = []
            for file in element.files:
                df = pd.read_csv(
                    join(self.downloaded_data_path, file.to_filename()),
                    index_col=None,
                    header=None,
                    names=["timestamp", "price"],
                )
                dfs.append(df)
            symbol = self.__get_data_symbol(element.symbol, element.source)
            symbol.additional_properties["chart_data_frame"] = pd.concat(dfs)

    async def __debug_breakpoint_event(self, breakpoint_params):
        breakpoint_params = DebugBreakpoint(**breakpoint_params)
        await self.__print_summary(
            breakpoint_params.custom_charts, breakpoint_params.display_charts
        )



from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import (
    CustomChartElement,
)
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.utils.config_validator import validate_config
from backtesterrb30.libs.utils.run_all import run_all_microservices
from backtesterrb30.libs.utils.strategy import Strategy
from backtesterrb30.python_engine.engine import Engine
from backtesterrb30.python_executor.trade_executor import Executor

__all__ = [
    "Executor",
    "Engine",
    "Strategy",
    "DataSchema",
    "CustomChartElement",
    "validate_config",
    "run_all_microservices",
    "HISTORICAL_SOURCES",
]



from backtesterrb30.libs.utils.run_service import run_service
from backtesterrb30.historical_data_feeds.historical_data_feeds import (
    HistoricalDataFeeds,
)

microservice_name = "historical_data_feeds"

run_service(microservice_name, HistoricalDataFeeds)



from backtesterrb30.libs.data_sources.binance.binance import BinanceDataSource
from backtesterrb30.libs.data_sources.coingecko.coingecko import CoingeckoDataSource
from backtesterrb30.libs.data_sources.dukascopy.dukascopy import DukascopyDataSource
from backtesterrb30.libs.data_sources.exante.exante import ExanteDataSource
from backtesterrb30.libs.data_sources.rb30.rb30_disk import RB30DataSource
from backtesterrb30.libs.data_sources.tradingview.tradingview import TradingView


class HISTORICAL_SOURCES:
    binance = BinanceDataSource
    dukascopy = DukascopyDataSource
    rb30disk = RB30DataSource
    exante = ExanteDataSource
    coingecko = CoingeckoDataSource
    tradingview = TradingView


# HISTORICAL_SOURCES_UNION = Union[BinanceDataSource, DukascopyDataSource, RB30DataSource, ExanteDataSource, CoingeckoDataSource]



import asyncio
from abc import abstractmethod
from typing import List, Union

from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.custom_chart import CustomChart
from backtesterrb30.libs.interfaces.python_backtester.data_finish import DataFinish
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import (
    DebugBreakpoint,
)
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import (
    CustomChartElement,
)
from backtesterrb30.libs.interfaces.python_engine.price_event import PriceEvent
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.json_serializable import JSONSerializable
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.module_loaders import (
    import_spec_module,
    reload_spec_module,
)
from backtesterrb30.libs.utils.service import Service


class Engine(Service):
    """Python Engine"""

    _broker: BrokerBase

    def __init__(
        self, config: Config, data_schema: DataSchema, loop=None, logger=print
    ):
        super().__init__(config, logger)
        self.config: Config = config
        self.__loop = loop
        self.__custom_event_loop = False
        if self.__loop is None:
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop = True
        self.__data_schema: DataSchema = data_schema
        self.__columns = ["timestamp"] + [c.symbol for c in self.__data_schema.data]
        self.__data_buffer = [[] for col in self.__columns]
        for sym, arr in zip(self.__data_schema.data, self.__data_buffer[1:]):
            sym.additional_properties["buffer"] = arr
            sym.additional_properties["price_events"] = []
        self.__buffer_length = 1
        self.__custom_charts: List[CustomChart] = []
        self.__debug_mode = False
        self.__debug_next_pressed = False
        self.__reloading_modules = []
        self.__send_breakpoint_after_feed = False
        self.__code_stopped_debug = False
        self.__breakpoint_display_charts = True

    # public methods:
    # ==================================================================

    @abstractmethod
    async def on_feed(self, data):
        """
        Function is being called when new data frame appears.
        While defining strategy you have to override this function
        and define your strategy logic here.

        :param data: data buffer defined in :class:`DataSchema`
        :type data: list
        """

    def on_data_finish(self):
        """Function is being called in backtest mode when historical data are finished"""

    def get_data_schema(self) -> DataSchema:
        """
        Returns data_schema defined in your strategy files in `data_schema.py`
        Data schema contains list of :class:`DataSymbol` objects. In every :class:`DataSymbol` object,
        you can use `get_buffer()` function to get current buffer connected to this symbol.

        :return: :class:`DataSchema` object combined with this strategy.
        :rtype: DataSchema
        """
        return self.__data_schema

    def get_data_symbol_by_custom_name(self, custom_name: str) -> DataSymbol:
        """
        Returns a :class:`DataSymbol` object using `custom_name` defined
        in `data_schema.py`, `custom_name` must be unique.

        :param custom_name: custom name defined in `data_schema.py` file.
        :type custom_name: str
        :return: :class:`DataSymbol` object combined with privided custom_name
        :rtype: DataSymbol
        """
        if not isinstance(custom_name, str):
            raise Exception("Provided name is not string")
        arr = [d for d in self.__data_schema.data if d.custom_name == custom_name]
        if len(arr) == 0:
            raise Exception("No data symbol with such custom name")
        if len(arr) > 1:
            raise Exception("Two elements with the same custom name")
        return arr[0]

    def get_columns(self) -> List[str]:
        """
        Returns a list of instrument names based on all instruments in :class:`DataSchema`.
        Do not use it as identifiers! Names can be the same in other data sources.

        :return: List of returned :class:`str` objects
        :rtype: list
        """
        return self.__columns

    def set_buffer_length(self, length: int):
        """
        Sets data buffer length. Data provided to `on_feed()` function
        is going to have this length.

        :param length: buffer length
        :type length: int
        """
        if length < 1:
            raise Exception("Buffer length must be bigget than 1")
        # if self.config.backtest == False and length > 1:
        #     raise Exception('Buffer longer than 1 not implemended in live mode')
        self.__buffer_length = length

    async def trigger_event(self, event: JSONSerializable):
        """
        Calling this function, you trigger `on_event()` function in
        :class:`TradeExecutor`. You can provide any JSON serializable input
        to this method.

        :param event: any JSON serializable data
        :type event: JSONSerializable
        """
        await self.__send_last_feed(SERVICES.python_executor)
        await self._broker.send(SERVICES.python_executor, "event", event)

    def add_custom_chart(
        self,
        chart: List[CustomChartElement],
        name: str,
        display_on_price_chart: Union[DataSymbol, None] = None,
        log_scale: Union[bool, None] = None,
        color: Union[str, None] = None,
    ):
        """
        Function adds custom chart printed under main charts in summary
        or in breakpoint debug event.

        :param chart: list of :class:`CustomChartElement`
        :type chart: list
        :param name: name of your custom name
        :type name: str
        :param display_on_price_chart: determining if chart is going to be displayed on other chart (on privided :class:`DataSymbol` chart) or it is going to be displayed alone
        :type display_on_price_chart: DataSymbol, optional
        :param log_scale: determines if chart is going to be in log scale
        :type log_scale: bool, optional
        :param color: a list of UUID strings, defaults to None
        :type color: str, optional
        """
        chart_obj = {
            "chart": chart,
            "display_on_price_chart": display_on_price_chart,
            "name": name,
        }
        if display_on_price_chart:
            chart_obj["display_on_price_chart"] = display_on_price_chart
        if log_scale:
            chart_obj["log_scale"] = log_scale
        if color:
            chart_obj["color"] = color
        self.__custom_charts.append(chart_obj)

    async def debug_breakpoint(self, display_charts=True):
        """
        This function causes breakpoint triggered while you turn on debug mode.

        :param display_charts: determines if breakpoint should cause displaying charts
        :type display_charts: bool, optional
        """
        self.__breakpoint_display_charts = display_charts
        if self.__debug_mode:
            self.__code_stopped_debug = True
            while True:
                if not self.__debug_mode:
                    for spec, module in self.__reloading_modules:
                        # reload(module)
                        reload_spec_module(spec, module)
                        # spec.loader.exec_module(module)

                        # module = SourceFileLoader('strategies.hackaton1.reloading_module.reloading_methods', '/home/george/workspace/project/Retire-Before-30/Engine-RB30/strategies/hackaton1/reloading_module/reloading_methods.py')

                    self.__code_stopped_debug = False
                    return
                if self.__debug_next_pressed:
                    for spec, module in self.__reloading_modules:
                        # reload(module)
                        reload_spec_module(spec, module)
                        spec.loader.exec_module(module)
                        # module = SourceFileLoader('strategies.hackaton1.reloading_module.reloading_methods', '/home/george/workspace/project/Retire-Before-30/Engine-RB30/strategies/hackaton1/reloading_module/reloading_methods.py')

                    self.__debug_next_pressed = False
                    self.__send_breakpoint_after_feed = True
                    self.__code_stopped_debug = False
                    return
                await asyncio.sleep(0.1)

    def add_reloading_module(self, module_path: str):
        """
        Adds module that is going to be reloaded
        every step of your debug. Module must not save any state
        because while realoading this state is going to be refreshed.

        :param module_path: path to module
        :type module_path: str
        """
        spec, module = import_spec_module(module_path)
        self.__reloading_modules.append((spec, module))
        reload_spec_module(spec, module)
        return module

    # ==================================================================
    # end of public methods

    # private methods:

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        if self.config.debug:
            self.__loop.create_task(self.__keyboard_listener())
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("data_feed", self.__data_feed_event)
        self._broker.register(
            "historical_sending_locked", self.__historical_sending_locked_event
        )
        self._broker.register("data_finish", self.__data_finish_event)
        self._broker.register("engine_ready", self.__engine_ready_event)
        self._broker.register("get_buffer_length", self.__get_buffer_length_event)
        self._broker.register("register_price_event", self.__register_price_event)

    # def _send(): pass
    # def _register(): pass
    # def _create_listeners(): pass

    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._broker._create_listeners(loop)
    #     loop.create_task(self.__keyboard_listener())

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    async def __send_last_feed(self, service: SERVICES):
        last_feed = {"last_feed": [v[-1] for v in self.__data_buffer]}
        await self._broker.send(service, "last_feed", LastFeed(**last_feed))

    async def __send_debug_breakpoint(self):
        breakpoint_params = {}
        breakpoint_params["custom_charts"] = self.__custom_charts
        breakpoint_params["display_charts"] = self.__breakpoint_display_charts
        self._log("sending debug breakpoint")
        await self.__send_last_feed(SERVICES.python_executor)
        await self._broker.send(
            SERVICES.python_executor,
            "debug_breakpoint",
            DebugBreakpoint(**breakpoint_params),
        )

    async def __keyboard_listener(self):
        import keyboard

        self._log('To enter debug mode press "ctrl+d"')
        while True:
            if keyboard.is_pressed("ctrl+d"):
                if not self.__debug_mode:
                    self._log(
                        'You have entered Debug mode \n\
                                 -> press "ctrl+n" to next step\n\
                                 -> press "ctrl+q" to leave debug mode'
                    )
                    self.__debug_mode = True
                while keyboard.is_pressed("d"):
                    await asyncio.sleep(0.1)
            if keyboard.is_pressed("ctrl+n"):
                if (
                    self.__debug_mode
                    and not self.__debug_next_pressed
                    and self.__code_stopped_debug
                ):
                    self._log(
                        'next step ... \n\
                                -> press "ctrl+q" to leave debug mode'
                    )
                    self.__debug_next_pressed = True
                while keyboard.is_pressed("n"):
                    await asyncio.sleep(0.1)
            if keyboard.is_pressed("ctrl+q"):
                if self.__debug_mode:
                    self._log(
                        'You have leaved Debug mode \n\
                                -> press "ctrl+d" to enter debug mode again'
                    )
                    self.__debug_mode = False
                while keyboard.is_pressed("q"):
                    await asyncio.sleep(0.1)
            await asyncio.sleep(0.1)

    # COMMANDS

    async def __data_feed_event(self, new_data_row):
        for i, v in enumerate(new_data_row):
            self.__data_buffer[i].append(v)
        if len(self.__data_buffer[0]) > self.__buffer_length:
            for i, v in enumerate(new_data_row):
                self.__data_buffer[i].pop(0)
        if len(self.__data_buffer[0]) >= self.__buffer_length:
            await self.on_feed(self.__data_buffer)
            if self.__send_breakpoint_after_feed:
                await self.__send_debug_breakpoint()
                self.__send_breakpoint_after_feed = False

    async def __historical_sending_locked_event(self):
        await self._broker.send(
            SERVICES.historical_data_feeds, "unlock_historical_sending"
        )

    async def __data_finish_event(self):
        self.on_data_finish()
        self.__debug_mode = False
        finish_params = {}
        finish_params["custom_charts"] = self.__custom_charts
        # finish_params['main_instrument_price'] = self.__get_main_intrument_price()
        await self.__send_last_feed(SERVICES.python_backtester)
        await self._broker.send(
            SERVICES.python_backtester, "data_finish", DataFinish(**finish_params)
        )

    async def __engine_ready_event(self, service_name):
        await self._broker.send(SERVICES[service_name], "engine_ready_response")

    async def __get_buffer_length_event(self, service_name):
        await self._broker.send(
            SERVICES[service_name], "engine_set_buffer_length", self.__buffer_length
        )

    async def __register_price_event(self, price_event):
        price_event = PriceEvent(**price_event)
        symbol: DataSymbol = [
            data
            for data in self.__data_schema.data
            if data.identifier == price_event.symbol_itentifier
        ][0]
        symbol.additional_properties["price_events"].append(price_event)



import asyncio
import os
import sys

from dotenv import load_dotenv

from backtesterrb30 import validate_config
from backtesterrb30.historical_data_feeds.historical_data_feeds import (
    HistoricalDataFeeds,
)
from backtesterrb30.libs.communication_broker.asyncio_broker import AsyncioBroker
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.live_data_feeds.live_data_feeds import LiveDataFeeds
from backtesterrb30.python_backtester.python_backtester import Backtester
from backtesterrb30.python_engine.engine import Engine
from backtesterrb30.python_executor.trade_executor import Executor
import platform

load_dotenv(".env")


class Strategy:
    def __init__(
        self, model: Engine, executor: Executor, data: dict, backtest=True, debug=False
    ):
        if debug and platform.system() != "Windows":
            if os.geteuid() != 0:
                print("You must be root to use debug mode because of keyboard package!")
                os._exit(1)
        if not backtest:
            print("Live is not yet supported")
            os._exit(1)
        self.__debug = debug
        self.__model = model
        self.__executor = executor
        self.__data = data

        self.__backtest_state = backtest
        self.__model_class_name = str(self.__model.__name__)
        self.__executor_class_name = str(self.__executor.__name__)
        self.__data_class_name = str(self.__data.__name__)
        splitted = os.path.normpath(sys.argv[0])
        splitted = splitted.split(os.sep)
        self.__strategy_path = os.path.join(str(os.getcwd()), *splitted[:-1])
        self.__strategy_file = splitted[-1]

        self.__loop = asyncio.get_event_loop()
        self.__services = {}
        self.__data_schema = validate_config(data.data)

    def create_service(
        self,
        microservice_name: str,
        service_class: Service,
        backtest_state,
        loop,
        data_schema,
    ) -> Service:
        config = {
            "name": microservice_name,
            "backtest": backtest_state,
            "strategy_path": self.__strategy_path,
            "debug": self.__debug,
        }
        service: Service = service_class(Config(**config), data_schema, loop)
        logger = service.get_logger()
        broker: AsyncioBroker = AsyncioBroker(service.config, logger)
        service.register_communication_broker(broker)
        return service, broker

    def run(self):
        static_params = self.__backtest_state, self.__loop, self.__data_schema
        self.__services["python_engine"] = self.create_service(
            "python_engine", self.__model, *static_params
        )
        self.__services["python_executor"] = self.create_service(
            "python_exacutor", self.__executor, *static_params
        )
        self.__services["historical_data_feeds"] = self.create_service(
            "historical_data_feeds", HistoricalDataFeeds, *static_params
        )
        self.__services["python_backtester"] = self.create_service(
            "python_backtester", Backtester, *static_params
        )
        self.__services["live_data_feeds"] = self.create_service(
            "live_data_feeds", LiveDataFeeds, *static_params
        )

        if self.__strategy_file != "serve.py":
            for name, (service, broker) in self.__services.items():
                service: Service = service
                broker: AsyncioBroker = broker
                for n, (s, b) in self.__services.items():
                    if n != name:
                        broker.register_broker(SERVICES[n], b)

            for name, (service, broker) in self.__services.items():
                service.run()
            self.__loop.run_forever()
            self.__loop.close()

    def run_in_microservices(self):
        if self.__strategy_file != "serve.py":
            from backtesterrb30 import run_all_microservices

            run_all_microservices(
                self.__strategy_path,
                self.__backtest_state,
                self.__strategy_file,
                self.__model_class_name,
                self.__executor_class_name,
                self.__data_class_name,
            )

    def run_in_docker_microservices():
        pass



from datetime import timezone

from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema


def validate_config(config: dict):
    for d in config["data"]:
        if not isinstance(d["historical_data_source"], str):
            d["historical_data_source"] = d["historical_data_source"].NAME
    cfg = DataSchema(**config)
    for symbol in cfg.data:
        symbol.backtest_date_start = symbol.backtest_date_start.replace(
            tzinfo=timezone.utc
        )
        symbol.backtest_date_stop = symbol.backtest_date_stop.replace(
            tzinfo=timezone.utc
        )
    return cfg
